@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (c) 2016-2017, 2019-2021 eGauge Systems LLC
+# Copyright (c) 2016-2017, 2019-2022 eGauge Systems LLC
 # 	1644 Conestoga St, Suite 2
 # 	Boulder, CO 80301
 # 	voice: 720-545-9767
@@ -35,7 +35,8 @@ import crcmod
 
 from .bit_stuffer import BitStuffer
 
-CTID_VERSION = 4  # version of CTid Specification this code implements
+
+CTID_VERSION = 5  # latest version of CTid Specification this code supports
 START_SYM = 0xFF
 
 SENSOR_TYPE_AC = 0x0  # AC-only sensor
@@ -86,19 +87,6 @@ SENSOR_UNITS = (
     ("", "number"),
 )
 
-_START_SYM_ZC_COUNT = 16  # number of zero-crossings in start-symbol
-_CLK_FREQ = 480  # nominal clock frequency of CTid carrier
-_CLK_TOL = 50  # clock tolerance in %
-
-_MAX_CLK_FREQ = math.ceil(_CLK_FREQ * (1.0 + (_CLK_TOL / 100)))
-_MIN_CLK_FREQ = math.floor(_CLK_FREQ * (1.0 - (_CLK_TOL / 100)))
-_MIN_CLK_PERIOD = 1e6 / _MAX_CLK_FREQ  # min. period in microseconds
-_MAX_CLK_PERIOD = 1e6 / _MIN_CLK_FREQ  # max. period in microseconds
-
-_START_SYM_MAX_DURATION = _START_SYM_ZC_COUNT / 2 * _MAX_CLK_PERIOD
-
-_DEBUG = 0
-
 crc8_rohc = crcmod.predefined.mkCrcFun("crc-8-rohc")
 crc16_modbus = crcmod.predefined.mkCrcFun("modbus")
 
@@ -113,7 +101,7 @@ class CRCError(Exception):
         self.got = got
 
     def __str__(self):
-        return "Expected CRC 0x%x but got 0x%x" % (self.expected, self.got)
+        return f"Expected CRC {self.expected:#x} but got {self.got:#x}"
 
 
 class Error(Exception):
@@ -123,7 +111,8 @@ class Error(Exception):
 def get_mfg_id_for_name(name):
     """Return the manufacturer ID for the given name.  Name may be a
     prefix of the full manufacturer's name, as long as the prefix is
-    unique.
+    unique.  Alternatively, name may also be a decimal integer string
+    or a hexdecimal string with prefix 0x.
 
     """
     match = None
@@ -132,9 +121,8 @@ def get_mfg_id_for_name(name):
             if value[0 : len(name)] == name:
                 if match:
                     raise Error(
-                        "Manufacturer name %s is a prefix of multiple "
-                        "registered names.  Please specify more "
-                        "letters." % name
+                        f"Manufacturer name {name} is a prefix of multiple "
+                        "registered names.  Please specify more letters."
                     )
                 match = key
     if match is not None:
@@ -170,226 +158,6 @@ def get_sensor_unit_desc(unit_code):
     return "?"
 
 
-def time_diff(l, r):
-    return l - r
-
-
-def is_zero_crossing(curr, prev, mean):
-    return (prev < mean) != (curr < mean)
-
-
-class Window:
-    def __init__(self, sampling_freq):
-        self.sample_period = 1e6 / sampling_freq  # period in microseconds
-        self.avg_len = math.ceil(_START_SYM_MAX_DURATION / self.sample_period)
-        window_len = 2 * self.avg_len
-
-        self.tolerance = None
-        self.count = 0  # number of valid samples
-        self.avg_count = 0  # number of averaged samples
-        self.sum = 0  # sum of the valid samples
-        self.mean = 0  # mean value of valid samples
-        self.wr = 0  # index of next sample to be written
-        self.prev_polarity = 0  # previous polarity
-        self.ts = [None] * window_len
-        self.val = [None] * window_len
-
-    def enter_sample(self, ts, val):
-        if self.avg_count < self.avg_len:
-            self.avg_count += 1
-        else:
-            idx = self.wr + self.avg_len
-            if idx >= len(self.val):
-                idx -= len(self.val)
-            self.sum -= self.val[idx]
-        if self.count < len(self.val):
-            self.count += 1
-        self.sum += val
-        self.val[self.wr] = val
-        self.ts[self.wr] = ts
-        self.wr = (self.wr + 1) % len(self.val)
-        self.mean = self.sum / self.avg_count
-
-    def at_zero_crossing(self):
-        curr_val = self.val[(self.wr + len(self.val) - 1) % len(self.val)]
-        curr_pol = curr_val >= self.mean
-        prev_pol = self.prev_polarity
-        self.prev_polarity = curr_pol
-        return curr_pol != prev_pol
-
-    def at_start_symbol(self):
-        if self.count < 1:
-            return None
-
-        curr = (self.wr + len(self.val) - 1) % len(self.val)
-        prev_val = self.val[curr]
-        self.prev_polarity = prev_val >= self.mean
-        end_ts = self.ts[curr]
-        start_ts = end_ts - _START_SYM_MAX_DURATION
-        total = 0
-        avg_count = 0
-
-        # See if we can find 16 zero-crossings within _START_SYM_MAX_DURATION:
-        zc_ts = []
-        for i in range(1, self.count):
-            curr = (curr + len(self.val) - 1) % len(self.val)
-            curr_ts = self.ts[curr]
-            curr_val = self.val[curr]
-
-            if time_diff(start_ts, curr_ts) > _MAX_CLK_PERIOD:
-                return None
-
-            total += curr_val
-            avg_count += 1
-
-            if not is_zero_crossing(curr_val, prev_val, self.mean):
-                continue  # keep looking for zero-crossing
-            prev_val = curr_val
-            zc_ts.append(curr_ts)
-            if len(zc_ts) >= _START_SYM_ZC_COUNT:
-                first_ts = curr_ts
-                break
-
-        if len(zc_ts) < _START_SYM_ZC_COUNT:
-            return None
-
-        mean = total / avg_count
-
-        # Verify that there were no zero-crossings for _START_SYM_MAX_DURATION
-        # before the first edge of the start-symbol
-        for i in range(i + 1, self.count):
-            curr = (curr + len(self.val) - 1) % len(self.val)
-            curr_ts = self.ts[curr]
-            curr_val = self.val[curr]
-
-            if is_zero_crossing(curr_val, prev_val, mean):
-                return None
-            if time_diff(first_ts, curr_ts) >= _START_SYM_MAX_DURATION:
-                break
-            prev_val = curr_val
-
-        # Calculate average period based on the 7 full bit times the 16
-        # zero-crossings cover (clock periods may be asymmetric):
-        period = time_diff(zc_ts[1], zc_ts[_START_SYM_ZC_COUNT - 1]) / (
-            _START_SYM_ZC_COUNT / 2 - 1
-        )
-        if period < _MIN_CLK_PERIOD or period > _MAX_CLK_PERIOD:
-            return None
-
-        self.tolerance = period / 4
-        sample_tolerance = 1.2 * self.sample_period
-        if self.tolerance < sample_tolerance:
-            self.tolerance = sample_tolerance
-
-        # Verify that each bit has the expected period:
-        curr_ts = end_ts
-        for i in range(0, _START_SYM_ZC_COUNT, 2):
-            dt = time_diff(curr_ts, zc_ts[i + 1])
-            if abs(dt - period) > self.tolerance:
-                return None
-            curr_ts = zc_ts[i + 1]
-
-        # Now that we confirmed a start-symbol, commit to its mean value:
-        self.sum = total
-        self.avg_count = avg_count
-        self.mean = mean
-        return period
-
-
-class ByteDecoder:
-    def __init__(self):
-        self.num_bits = 0
-        self.val = 0
-        self.edge_count = 0
-        self.run_length = 0
-        self.start_ts = None
-        self.decoded_byte = None
-
-    def reset(self):
-        self.__init__()
-
-    def timed_out(self, period, now):
-        if self.start_ts is None:
-            return False
-        return time_diff(now, self.start_ts) > 4 * period
-
-    def update(self, period, tolerance, ts):
-        if self.start_ts is None:
-            self.start_ts = ts
-            return False
-
-        if time_diff(ts, self.start_ts + period - tolerance) < 0:
-            # got what looks like a 1-edge
-            self.edge_count += 1
-            return False
-
-        bit = self.edge_count & 1
-        if _DEBUG >= 2:
-            print(
-                "bit %d at 0x%08x (edge_count %d, start_ts 0x%08x)"
-                % (bit, ts, self.edge_count, self.start_ts)
-            )
-
-        self.edge_count = 0
-        self.start_ts = ts
-        if self.run_length >= 7:
-            # drop stuffer bit...
-            self.run_length = 0
-            return False
-
-        if bit:
-            self.run_length += 1
-        else:
-            self.run_length = 0
-
-        self.val = (self.val << 1) | bit
-        self.num_bits += 1
-
-        if self.num_bits < 8:
-            return False
-        self.decoded_byte = self.val
-        self.val = 0
-        self.num_bits = 0
-        return True
-
-    def get_byte(self):
-        return self.decoded_byte
-
-
-class Decoder:
-    def __init__(self, sampling_freq):
-        self.w = Window(sampling_freq)
-        self.bd = ByteDecoder()
-        self.period = None
-
-    def get_byte(self):
-        return self.bd.get_byte()
-
-    def add_sample(self, timestamp, value):
-        self.w.enter_sample(timestamp, value)
-
-        if self.period is None:
-            self.period = self.w.at_start_symbol()
-            if self.period is None:
-                return 0
-            self.bd.reset()
-            return -1  # got a new start-symbol
-
-        if self.bd.timed_out(self.period, timestamp):
-            # check if implied last edge finishes a byte, if so, return it:
-            ret = self.bd.update(self.period, self.w.tolerance, timestamp)
-            self.period = None
-            return ret
-
-        if not self.w.at_zero_crossing():
-            return 0
-        if _DEBUG >= 3:
-            print(
-                "zero-crossing at 0x%08x (mean %d)" % (timestamp, self.w.mean)
-            )
-        return self.bd.update(self.period, self.w.tolerance, timestamp)
-
-
 def mfg_name(ident):
     """Return the manufacturer name for a given manufacturer IDENT."""
     if ident in _MFG_ID:
@@ -409,20 +177,29 @@ def get_mfg_id(ident):
     return mfg_name(ident)
 
 
+def s10(val):
+    """Convert 10 bit unsigned integer to a signed 9-bit integer."""
+    if not 0 <= val < 1024:
+        raise ValueError
+    if val < 512:
+        return val
+    return val - 1024
+
+
 def fix(val, num_bits, is_signed, name, unit, scale):
     f = round(val / scale)
     limit = 1 << num_bits
     min_val = -(limit / 2) if is_signed else 0
     max_val = +(limit / 2 - 1) if is_signed else limit - 1
     if f < min_val or f > max_val:
-        unit_str = " %s" % unit if unit is not None else ""
+        unit_str = (" " + unit) if unit is not None else ""
         prec = -int(math.log10(scale))
-        val_str = ("%." + str(prec) + "f") % (val)
-        min_str = ("%." + str(prec) + "f") % ((min_val * scale))
-        max_str = ("%." + str(prec) + "f") % ((max_val * scale))
+        val_str = f"{val:.{prec}f}"
+        min_str = f"{min_val * scale:.{prec}f}"
+        max_str = f"{max_val * scale:.{prec}f}"
         raise Error(
-            "%s %s%s outside of range from %s to %s%s."
-            % (name, val_str, unit_str, min_str, max_str, unit_str)
+            f"{name} {val_str}{unit_str} outside of range "
+            f"from {min_str} to {max_str}{unit_str}."
         )
     return f
 
@@ -438,8 +215,7 @@ def float12(val, name, unit):
     """
     if val < 0 or val > 1e7:
         raise Error(
-            "%s %s%s outside of range from 0 to 10,000,000%s"
-            % (name, val, unit, unit)
+            f"{name} {val}{unit} outside of range from 0 to 10,000,000{unit}"
         )
     exp = math.log10(val) if val >= 1.0 else 0
     exp = int(math.ceil(exp))
@@ -478,8 +254,8 @@ def decimal16(val, name, unit):
             exp += 1
         if exp > 15:
             raise Error(
-                "%s %s%s outside of range from -1024e15..1023e15%s"
-                % (name, val, unit, unit)
+                f"{name} {val}{unit} outside of range "
+                f"from -1024e15..1023e15{unit}"
             )
         mantissa = int(mantissa)
     return ((mantissa & 0x7FF) << 5) | (exp & 0x1F)
@@ -584,83 +360,63 @@ class Table:
         if mfg_id_str is not None:
             mfg_id_str = '"' + mfg_id_str + '"'
         else:
-            mfg_id_str = "0x%04x" % self.mfg_id
+            mfg_id_str = f"{self.mfg_id:#04x}"
 
         sensor_type_str = get_sensor_type_name(self.sensor_type)
         if sensor_type_str is not None:
             sensor_type_str = '"' + sensor_type_str + '"'
         else:
-            sensor_type_str = "0x%1x" % self.sensor_type
+            sensor_type_str = f"{self.sensor_type:#1x}"
 
         ret = (
-            "CTid {version=%d, sensor_type=%s, mfg_id=%s, "
-            'model="%s", '
-            % (self.version, sensor_type_str, mfg_id_str, self.model)
+            f"CTid version={self.version:d}, sensor_type={sensor_type_str}, "
+            f'mfg_id={mfg_id_str}, model="{self.model}", '
         )
         if (
             self.sensor_type >= SENSOR_TYPE_AC
             and self.sensor_type <= SENSOR_TYPE_RC
         ):
+            rows = []
+            for lvl, val in self.cal_table.items():
+                rows.append(f"{repr(lvl)}%: {val[0]:+.2f}%/{val[1]:+.2f}°")
             ret += (
-                "size=%.1fmm, serial=%u, current=%.1fA, voltage=%.6fV, "
-                "bias=%.3fmV, "
-                "phase=%.2f\u00b0, voltage_temp_coeff=%.0fppm/\u00b0C, "
-                "phase_temp_coeff=%.1fm\u00b0/\u00b0C, reserved=0x%02x, "
-                "mfg_info=0x%02x, Rs=%g, Rl=%g, "
-                "cal={"
-                "1.5%%: %+.2f%%/%+.2f\u00b0, "
-                "5%%: %+.2f%%/%+.2f\u00b0, "
-                "15%%: %+.2f%%/%+.2f\u00b0, "
-                "50%%: %+.2f%%/%+.2f\u00b0}"
-                % (
-                    self.size,
-                    self.serial_number,
-                    self.rated_current,
-                    self.voltage_at_rated_current,
-                    1e3 * self.bias_voltage,
-                    self.phase_at_rated_current,
-                    self.voltage_temp_coeff,
-                    self.phase_temp_coeff,
-                    self.reserved,
-                    self.mfg_info,
-                    self.r_source,
-                    self.r_load,
-                    self.cal_table[1.5][0],
-                    self.cal_table[1.5][1],
-                    self.cal_table[5][0],
-                    self.cal_table[5][1],
-                    self.cal_table[15][0],
-                    self.cal_table[15][1],
-                    self.cal_table[50][0],
-                    self.cal_table[50][1],
-                )
+                f"size={self.size:.1f}mm, "
+                f"serial={self.serial_number}, "
+                f"current={self.rated_current:.1f}A, "
+                f"voltage={self.voltage_at_rated_current:.6f}V, "
+                f"bias={1e3 * self.bias_voltage:.3f}mV, "
+                f"phase={self.phase_at_rated_current:.2f}°, "
+                f"voltage_temp_coeff={self.voltage_temp_coeff:.0f}ppm/°C, "
+                f"phase_temp_coeff={self.phase_temp_coeff:.1f}m°/°C, "
+                f"reserved={self.reserved:#02x}, "
+                f"mfg_info={self.mfg_info:#02x}, "
+                f"Rs={self.r_source:g}, "
+                f"Rl={self.r_load:g}, "
+                "cal={" + ", ".join(rows) + "}"
             )
         elif self.sensor_type == SENSOR_TYPE_LINEAR:
             unit = get_sensor_unit(self.sensor_unit)
-            ret += "scale=%g%s/V offset=%g%s delay=%gμs" % (
-                self.scale,
-                unit,
-                self.offset,
-                unit,
-                self.delay,
+            ret += (
+                f"scale={self.scale:g}{unit}/V "
+                f"offset={self.offset:g}{unit} "
+                f"delay={self.delay:g}μs"
             )
         elif self.sensor_type == SENSOR_TYPE_TEMP_LINEAR:
-            ret += "scale=%g°C/V offset=%g°C" % (self.scale, self.offset)
+            ret += f"scale={self.scale:g}°C/V " f"offset={self.offset:g}°C"
         elif self.sensor_type == SENSOR_TYPE_TEMP_NTC:
-            ret += "A=%g B=%g C=%g M=%g N=%g K=%g" % (
-                self.ntc_a,
-                self.ntc_b,
-                self.ntc_c,
-                self.ntc_m,
-                self.ntc_n,
-                self.ntc_k,
+            ret += (
+                f"A={self.ntc_a:g} "
+                f"B={self.ntc_b:g} "
+                f"C={self.ntc_c:g} "
+                f"M={self.ntc_m:g} "
+                f"N={self.ntc_n:g} "
+                f"K={self.ntc_k:g}"
             )
         elif self.sensor_type == SENSOR_TYPE_PULSE:
-            ret += "threshold=%g±%gV debounce=%ums edge=0x%x" % (
-                self.threshold,
-                self.hysteresis,
-                self.debounce_time,
-                self.edge_mask,
+            ret += (
+                f"threshold={self.threshold:g}±{self.hysteresis:g}V "
+                f"debounce={self.debounce_time:d}ms "
+                f"edge={self.edge_mask:#x}"
             )
         else:
             pass
@@ -670,75 +426,105 @@ class Table:
     def m_u2(self, name, scale=1, unit=None):
         fmt = ">B"
         if self.encoding:
-            val = fix(self.__dict__[name], 2, False, name, unit, scale)
+            val = fix(getattr(self, name), 2, False, name, unit, scale)
             self._raw_data += struct.pack(fmt, val << 6)
         else:
             val = struct.unpack_from(fmt, self._raw_data, self._raw_offset)[0]
             self._raw_offset += struct.calcsize(fmt)
-            self.__dict__[name] = unfix(val >> 6, scale)
+            setattr(self, name, unfix(val >> 6, scale))
 
     def m_u8(self, name, scale=1, unit=None):
         fmt = ">B"
         if self.encoding:
-            val = fix(self.__dict__[name], 8, False, name, unit, scale)
+            val = fix(getattr(self, name), 8, False, name, unit, scale)
             self._raw_data += struct.pack(fmt, val)
         else:
             val = struct.unpack_from(fmt, self._raw_data, self._raw_offset)[0]
             self._raw_offset += struct.calcsize(fmt)
-            self.__dict__[name] = unfix(val, scale)
+            setattr(self, name, unfix(val, scale))
 
     def m_s8(self, name, scale=1, unit=None, idx1=None, idx2=None):
         fmt = ">b"
         if self.encoding:
             if idx1 is not None and idx2 is not None:
                 val = fix(
-                    self.__dict__[name][idx1][idx2], 8, True, name, unit, scale
+                    getattr(self, name)[idx1][idx2], 8, True, name, unit, scale
                 )
             else:
-                val = fix(self.__dict__[name], 8, True, name, unit, scale)
+                val = fix(getattr(self, name), 8, True, name, unit, scale)
             self._raw_data += struct.pack(fmt, val)
         else:
             val = struct.unpack_from(fmt, self._raw_data, self._raw_offset)[0]
             self._raw_offset += struct.calcsize(fmt)
             if idx1 is not None and idx2 is not None:
-                self.__dict__[name][idx1][idx2] = unfix(val, scale)
+                t = getattr(self, name)
+                t[idx1][idx2] = unfix(val, scale)
             else:
-                self.__dict__[name] = unfix(val, scale)
+                setattr(self, name, unfix(val, scale))
+
+    def m_4s10(self, name, levels):
+        """Pack/unpack four signed 10-bit words into/from 5 unsigned bytes."""
+        fmt = 5 * "B"
+        t = getattr(self, name)
+        if self.encoding:
+            v0 = fix(t[levels[0]][0], 10, True, name, "%", 0.01)
+            v1 = fix(t[levels[0]][1], 10, True, name, "°", 0.01)
+            v2 = fix(t[levels[1]][0], 10, True, name, "%", 0.01)
+            v3 = fix(t[levels[1]][1], 10, True, name, "°", 0.01)
+            bs = (
+                (v0 >> 2) & 0xff,
+                ((v0 & 0x03) << 6) | ((v1 >> 4) & 0x3f),
+                ((v1 & 0x0f) << 4) | ((v2 >> 6) & 0x0f),
+                ((v2 & 0x3f) << 2) | ((v3 >> 8) & 0x03),
+                (v3 & 0xff),
+            )
+            self._raw_data += struct.pack(fmt, *bs)
+        else:
+            bs = struct.unpack_from(fmt, self._raw_data, self._raw_offset)
+            self._raw_offset += struct.calcsize(fmt)
+            v0 = s10((bs[0] << 2) | ((bs[1] >> 6) & 0x03))
+            v1 = s10(((bs[1] & 0x3f) << 4) | ((bs[2] >> 4) & 0x0f))
+            v2 = s10(((bs[2] & 0x0f) << 6) | ((bs[3] >> 2) & 0x3f))
+            v3 = s10(((bs[3] & 0x03) << 8) | bs[4])
+            t[levels[0]][0] = unfix(v0, 0.01)
+            t[levels[0]][1] = unfix(v1, 0.01)
+            t[levels[1]][0] = unfix(v2, 0.01)
+            t[levels[1]][1] = unfix(v3, 0.01)
 
     def m_u16(self, name, scale=1, unit=None):
         fmt = ">H"
         if self.encoding:
-            val = fix(self.__dict__[name], 16, False, name, unit, scale)
+            val = fix(getattr(self, name), 16, False, name, unit, scale)
             self._raw_data += struct.pack(fmt, val)
         else:
             val = struct.unpack_from(fmt, self._raw_data, self._raw_offset)[0]
             self._raw_offset += struct.calcsize(fmt)
-            self.__dict__[name] = unfix(val, scale)
+            setattr(self, name, unfix(val, scale))
 
     def m_s16(self, name, scale=1, unit=None):
         fmt = ">h"
         if self.encoding:
-            val = fix(self.__dict__[name], 16, True, name, unit, scale)
+            val = fix(getattr(self, name), 16, True, name, unit, scale)
             self._raw_data += struct.pack(fmt, val)
         else:
             val = struct.unpack_from(fmt, self._raw_data, self._raw_offset)[0]
             self._raw_offset += struct.calcsize(fmt)
-            self.__dict__[name] = unfix(val, scale)
+            setattr(self, name, unfix(val, scale))
 
     def m_d16(self, name, unit):
         fmt = ">H"
         if self.encoding:
-            val = self.__dict__[name]
+            val = getattr(self, name)
             self._raw_data += struct.pack(fmt, decimal16(val, name, unit))
         else:
             val = struct.unpack_from(fmt, self._raw_data, self._raw_offset)[0]
             self._raw_offset += struct.calcsize(fmt)
-            self.__dict__[name] = undecimal16(val)
+            setattr(self, name, undecimal16(val))
 
     def m_s12(self, name, scale=1, unit=None):
         fmt = ">H"
         if self.encoding:
-            s12 = fix(self.__dict__[name], 12, True, name, unit, scale)
+            s12 = fix(getattr(self, name), 12, True, name, unit, scale)
             if s12 < 0:
                 s12 &= 0xFFF
             self._raw_data += struct.pack(fmt, s12 << 4)
@@ -748,43 +534,43 @@ class Table:
             s12 = (u16 >> 4) & 0x0FFF
             if s12 >= 0x800:
                 s12 -= 0x1000
-            self.__dict__[name] = unfix(s12, scale)
+            setattr(self, name, unfix(s12, scale))
 
     def m_u4_s12(self, name4, scale4, unit4, name12, scale12=1, unit12=None):
         fmt = ">H"
         if self.encoding:
-            u4 = fix(self.__dict__[name4], 4, False, name4, unit4, scale4)
-            s12 = fix(self.__dict__[name12], 12, True, name12, unit12, scale12)
+            u4 = fix(getattr(self, name4), 4, False, name4, unit4, scale4)
+            s12 = fix(getattr(self, name12), 12, True, name12, unit12, scale12)
             if s12 < 0:
                 s12 &= 0xFFF
             self._raw_data += struct.pack(fmt, (u4 << 12) | s12)
         else:
             u16 = struct.unpack_from(fmt, self._raw_data, self._raw_offset)[0]
             self._raw_offset += struct.calcsize(fmt)
-            self.__dict__[name4] = unfix((u16 >> 12) & 0xF, scale4)
+            setattr(self, name4, unfix((u16 >> 12) & 0xF, scale4))
             s12 = u16 & 0x0FFF
             if s12 >= 0x800:
                 s12 -= 0x1000
-            self.__dict__[name12] = unfix(s12, scale12)
+            setattr(self, name12, unfix(s12, scale12))
 
     def m_f12_f12(self, name1, unit1, name2=1, unit2=None):
         fmt = ">BH"
         if self.encoding:
-            f1 = float12(self.__dict__[name1], name1, unit1)
-            f2 = float12(self.__dict__[name2], name2, unit2)
+            f1 = float12(getattr(self, name1), name1, unit1)
+            f2 = float12(getattr(self, name2), name2, unit2)
             u24 = (f1 << 12) | f2
             self._raw_data += struct.pack(fmt, u24 >> 16, u24 & 0xFFFF)
         else:
             val = struct.unpack_from(fmt, self._raw_data, self._raw_offset)
             u24 = (val[0] << 16) | val[1]
             self._raw_offset += struct.calcsize(fmt)
-            self.__dict__[name1] = round(unfloat12((u24 >> 12) & 0xFFF))
-            self.__dict__[name2] = round(unfloat12((u24 >> 0) & 0xFFF))
+            setattr(self, name1, round(unfloat12((u24 >> 12) & 0xFFF)))
+            setattr(self, name2, round(unfloat12((u24 >> 0) & 0xFFF)))
 
     def m_u24(self, name, scale=1, unit=None):
         fmt = ">BH"
         if self.encoding:
-            val = fix(self.__dict__[name], 24, False, name, unit, scale)
+            val = fix(getattr(self, name), 24, False, name, unit, scale)
             self._raw_data += struct.pack(
                 fmt, (val >> 16) & 0xFF, val & 0xFFFF
             )
@@ -792,25 +578,24 @@ class Table:
             val = struct.unpack_from(fmt, self._raw_data, self._raw_offset)
             val = (val[0] << 16) | val[1]
             self._raw_offset += struct.calcsize(fmt)
-            self.__dict__[name] = unfix(val, scale)
+            setattr(self, name, unfix(val, scale))
 
     def m_f32(self, name):
         fmt = ">f"
         if self.encoding:
-            self._raw_data += struct.pack(fmt, self.__dict__[name])
+            self._raw_data += struct.pack(fmt, getattr(self, name))
         else:
             val = struct.unpack_from(fmt, self._raw_data, self._raw_offset)
             self._raw_offset += struct.calcsize(fmt)
-            self.__dict__[name] = val[0]
+            setattr(self, name, val[0])
 
     def m_utf8_4(self, name):
         if self.encoding:
-            val = self.__dict__[name].encode("utf-8")
+            val = getattr(self, name).encode("utf-8")
             if len(val) > 4:
                 raise Error(
-                    "%s `%s' is %u bytes long in UTF-8 "
-                    "but only up to 4 bytes are allowed."
-                    % (name, self.__dict__[name], len(val))
+                    f"{name} `{getattr(self, name)}' is {len(val)} bytes "
+                    "long in UTF-8 but only up to 4 bytes are allowed."
                 )
             while len(val) < 4:
                 val += b"\0"
@@ -822,16 +607,15 @@ class Table:
                     break
                 raw += bytearray((self._raw_data[i],))
             self._raw_offset += 4
-            self.__dict__[name] = raw.decode()
+            setattr(self, name, raw.decode())
 
     def m_utf8_8(self, name):
         if self.encoding:
-            val = self.__dict__[name].encode("utf-8")
+            val = getattr(self, name).encode("utf-8")
             if len(val) > 8:
                 raise Error(
-                    "%s `%s' is %u bytes long in UTF-8 "
-                    "but only up to 8 bytes are allowed."
-                    % (name, self.__dict__[name], len(val))
+                    f"{name} `{getattr(self, name)}' is {len(val)} bytes "
+                    "long in UTF-8 but only up to 8 bytes are allowed."
                 )
             while len(val) < 8:
                 val += b"\0"
@@ -843,23 +627,31 @@ class Table:
                     break
                 raw += bytearray((self._raw_data[i],))
             self._raw_offset += 8
-            self.__dict__[name] = raw.decode()
+            setattr(self, name, raw.decode())
 
     def marshall_params_ct(self):
         """Current Transducer Parameters"""
         self.m_u16("rated_current", 0.1, "A")
         self.m_u16("voltage_at_rated_current", 10e-6, "V")
         self.m_u16("size", 0.1, "mm")
-        self.m_s12("phase_at_rated_current", 0.01, "\u00b0")
-        self.m_s8("voltage_temp_coeff", 5, "ppm/\u00b0C")
-        self.m_s8("phase_temp_coeff", 0.5, "m\u00b0/\u00b0C")
-        for k in sorted(self.cal_table.keys()):
-            self.m_s8("cal_table", 0.02, "%", idx1=k, idx2=0)
-            self.m_s8("cal_table", 0.02, "\u00b0", idx1=k, idx2=1)
+        self.m_s12("phase_at_rated_current", 0.01, "°")
+        self.m_s8("voltage_temp_coeff", 5, "ppm/°C")
+        self.m_s8("phase_temp_coeff", 0.5, "m°/°C")
+        if self.version >= 5:
+            keys = list(self.cal_table.keys())
+            if len(keys) != 4:
+                raise Error(f"Cal table has {len(keys)} rows; expected 4.")
+            self.m_4s10("cal_table", keys[0:2])
+            self.m_4s10("cal_table", keys[2:4])
+        else:
+            for k in sorted(self.cal_table.keys()):
+                self.m_s8("cal_table", 0.02, "%", idx1=k, idx2=0)
+                self.m_s8("cal_table", 0.02, "\u00b0", idx1=k, idx2=1)
         self.m_s16("bias_voltage", 1e-6, "V")
 
     def marshall_params_basic_linear(self):
-        """Basic linear Parameters (for linear temp. and generic linear sensors)."""
+        """Basic linear Parameters (for linear temp. and generic linear
+        sensors)."""
         self.m_f32("scale")
         self.m_f32("offset")
 
@@ -942,19 +734,22 @@ class Table:
                     None,
                     "phase_at_rated_current",
                     0.01,
-                    "\u00b0",
+                    "°",
                 )
-                self.m_s8("voltage_temp_coeff", 5, "ppm/\u00b0C")
-                self.m_s8("phase_temp_coeff", 0.5, "m\u00b0/\u00b0C")
+                self.m_s8("voltage_temp_coeff", 5, "ppm/°C")
+                self.m_s8("phase_temp_coeff", 0.5, "m°/°C")
                 for k in sorted(self.cal_table.keys()):
                     self.m_s8("cal_table", 0.02, "%", idx1=k, idx2=0)
-                    self.m_s8("cal_table", 0.02, "\u00b0", idx1=k, idx2=1)
+                    self.m_s8("cal_table", 0.02, "°", idx1=k, idx2=1)
                 self.m_u8("reserved")
                 self.m_u8("mfg_info")
                 if self.version > 1:
                     self.m_f12_f12("r_source", "Ω", "r_load", "Ω")
 
-    def encode(self, version=CTID_VERSION):
+    # TODO: Default to version 4 for now.  Once eGauge firmware v4.1.3
+    # or later is widely in use, we can switch this back to
+    # version=CTID_VERSION.
+    def encode(self, version=4):
         """Encode the table contents and store is as a sequence of bytes in
         property ``raw_data''.
 
@@ -966,7 +761,7 @@ class Table:
         if version < 2 and (self.r_source != 0.0 or self.r_load != 0.0):
             raise Error(
                 "Unable to encode r_source and r_load in "
-                "CTid table version %u." % version
+                f"CTid table version {version}."
             )
         try:
             self.marshall()
